
#include "max7221.h"

volatile uint8_t  ind[8];  // регистры индикации разрядов
volatile uint8_t indb[8]; // регистры буфера индикации
//======================================================================================
// символы знакогенератора
const static uint8_t sg[] = {
     0b01111110,      // 0
     0b00110000,      // 1
     0b01101101,      // 2
     0b01111001,      // 3
     0b00110011,      // 4
     0b01011011,      // 5
     0b01011111,      // 6
     0b01110000,      // 7
     0b01111111,      // 8
     0b01111011,      // 9
     0b01100011,      // символ "°" градус Цельсия
     0b01001110,      // "C"
};
//======================================================================================

uint8_t i, j, pomt;              // переменные для эффектов

/* конструктор */
MAX7221::MAX7221(uint8_t dataPin, uint8_t clkPin, uint8_t csPin) { 
  SPI_MOSI = dataPin;
  SPI_SCK = clkPin;
  SPI_CS = csPin;
  pinMode(SPI_MOSI, OUTPUT);
  pinMode(SPI_SCK, OUTPUT);
  pinMode(SPI_CS, OUTPUT);
  digitalWrite(SPI_CS, HIGH); // высокий уровень

  // режим работы
  delayMicroseconds(2);
  spi_go();
  spi_transfer(0x09);
  spi_transfer(0x00);
  spi_end();
  delayMicroseconds(2);

  // тест дисплея выключен
  spi_go();
  spi_transfer(0x0F);
  spi_transfer(0X00);
  spi_end();
  delayMicroseconds(2);

  // включение дисплея
  spi_go();
  spi_transfer(0x0C);
  spi_transfer(0x01);
  spi_end();
  delayMicroseconds(2);

  // Яркость дисплея
  spi_go();
  spi_transfer(0x0A);
  spi_transfer(8);
  spi_end();
  delayMicroseconds(2);

  // количество сканируемых разрядов в драйвере
  spi_go();
  spi_transfer(0x0B);
  spi_transfer(7);
  spi_end();
  delayMicroseconds(2);

  clear_DISPL();
}

/* Реализация программного SPI режима */
void MAX7221::spi_transfer(uint8_t data) {
  for (i = 0; i < 8; i++) {                 // цикл на 8 бит
    delayMicroseconds(SPI_time/2);      // формируем половину паузы
    if (data & 0x80)
      digitalWrite(SPI_MOSI, HIGH);           // высокий уровень
    else
      digitalWrite(SPI_MOSI, LOW);           // низкий уровень
    data <<= 1;                            // двигаем байт
    delayMicroseconds(SPI_time/2);      // завершаем паузу
    digitalWrite(SPI_SCK, HIGH);           // тактовый фронт
    delayMicroseconds(SPI_time);          // формируем импульс
    digitalWrite(SPI_SCK, LOW);            // тактовый спад
  }
}
/* Функция передачи двух пакетов по 8 бит по протоколу SPI */
void MAX7221::LOAD_displ(void) {           
  for (j = 0; j < 8; j++) {
    spi_go();
    spi_transfer(j + 1);
    spi_transfer(ind[j]);
    spi_end();
  }
}

/* Очистка дисплея, все биты индикаци сбрасывает в ноль */
void MAX7221::clear_DISPL (void) { 
  for (j = 1; j < 9; j++) {
    spi_go();
    spi_transfer(j);
    spi_transfer(0x00);
    spi_end();
    delayMicroseconds(2);
  }
}
/* яркость дисплея */
void MAX7221::yark_DISPL (char yarkost) { 
  spi_go();
  spi_transfer(0x0A);
  spi_transfer(yarkost);
  spi_end();
  delayMicroseconds(2);
}

/* начинаем общение */
void MAX7221::spi_go() {
  delayMicroseconds(SPI_time);
  digitalWrite(SPI_SCK, LOW); // низкий уровень
  digitalWrite(SPI_CS, LOW); // низкий уровень
  delayMicroseconds(SPI_time);
}
/* заканчиваем общение */
void MAX7221::spi_end() {
  delayMicroseconds(SPI_time);
  digitalWrite(SPI_CS, HIGH); // высокий уровень
  digitalWrite(SPI_SCK, LOW); // низкий уровень
  delayMicroseconds(SPI_time);
}

/* загрузка буфера индикатора и непосредственная индикация */
void MAX7221::indic (uint8_t buf1, uint8_t buf2, uint8_t buf3, uint8_t buf4, boolean dot2, boolean dot4) {
  ind[0] = sg[buf1];
  ind[2] = sg[buf2];
  ind[4] = sg[buf3];
  ind[6] = sg[buf4];
  if(dot2) ind[2] |= 0b10000000;
  else ind[2] |= 0b00000000;
  if(dot4) ind[6] |= 0b10000000;
  else ind[6] |= 0b00000000;

  LOAD_displ();
}
/* загрузка буфера (для эффектов анимации индикатора) */
void MAX7221::LoadBU(uint8_t buf1, uint8_t buf2, uint8_t buf3, uint8_t buf4, boolean dot2, boolean dot4) {
  indb[0] = sg[buf1];
  indb[2] = sg[buf2];
  indb[4] = sg[buf3];
  indb[6] = sg[buf4];
  if(dot2) ind[2] |= 0b10000000;
  else ind[2] |= 0b00000000;
  if(dot4) ind[6] |= 0b10000000;
  else ind[6] |= 0b00000000;
}
/* задежка для создания эффектов индикации */
void MAX7221::zaderj(uint8_t dlit)
{
  uint8_t pous;
  for (pous = 0; pous < dlit; pous++)
    delay(10);
}

/* очистка индикатора слева направо */
void MAX7221::ochisL (uint8_t pomt)
{
  uint8_t cikle;
  for (cikle = 0; cikle < 4; cikle++)
  {
    ind[0] = ind[2];
    ind[2] = ind[4];
    ind[4] = ind[6];
    ind[6] = 0;
    LOAD_displ();
    zaderj(pomt);
  }
}

/* очистка индикатора справа налево */
void MAX7221::ochisR (uint8_t pomt)
{
  uint8_t cikle;
  for (cikle = 0; cikle < 8; cikle++)
  {
    ind[0] = ind[2];
    ind[2] = ind[4];
    ind[4] = ind[6];
    ind[6] = 0;
    LOAD_displ();
    zaderj(pomt);
  }
}
/* выезд буфера справо налево */
void MAX7221::go_left(void)
{
  uint8_t cikle;
  ochisL(DLIT);

  for (cikle = 0; cikle < 8; cikle++)
  {

    ind[0] = ind[2];
    ind[2] = ind[4];
    ind[4] = ind[6];
    ind[6] = indb[cikle];
    LOAD_displ();
    zaderj(DLIT);
  }
}
/* выезд буфера слево направо */
void MAX7221::go_right(void)
{
  uint8_t cikle;
  ochisR(DLIT);
  for (cikle = 0; cikle < 8; cikle++)
  {
    ind[6] = ind[4];
    ind[4] = ind[2];
    ind[2] = ind[0];
    ind[0] = indb[7 - cikle];
    LOAD_displ();
    zaderj(DLIT);
  }
}
/* выезд буфера сверху вниз */
void MAX7221::go_down(void)
{
  //очистка
  ind[0] &= 0b01100011;
  ind[2] &= 0b01100011;
  ind[4] &= 0b01100011;
  ind[6] &= 0b01100011;

  if (0b00000001 & ind[0]) ind[0] |= 0b00001000;
  if (0b00000001 & ind[2]) ind[2] |= 0b00001000;
  if (0b00000001 & ind[4]) ind[4] |= 0b00001000;
  if (0b00000001 & ind[6]) ind[6] |= 0b00001000;

  ind[0] &= 0b01111110;
  ind[2] &= 0b01111110;
  ind[4] &= 0b01111110;
  ind[6] &= 0b01111110;

  if (0b00100000 & ind[0]) ind[0] |= 0b00010000;
  if (0b00100000 & ind[2]) ind[2] |= 0b00010000;
  if (0b00100000 & ind[4]) ind[4] |= 0b00010000;
  if (0b00100000 & ind[6]) ind[6] |= 0b00010000;

  if (0b00000010 & ind[0]) ind[0] |= 0b00000100;
  if (0b00000010 & ind[2]) ind[2] |= 0b00000100;
  if (0b00000010 & ind[4]) ind[4] |= 0b00000100;
  if (0b00000010 & ind[6]) ind[6] |= 0b00000100;

  if (0b01000000 & ind[0]) ind[0] |= 0b00000001;
  if (0b01000000 & ind[2]) ind[2] |= 0b00000001;
  if (0b01000000 & ind[4]) ind[4] |= 0b00000001;
  if (0b01000000 & ind[6]) ind[6] |= 0b00000001;

  ind[0] &= 0b00011101;
  ind[2] &= 0b00011101;
  ind[4] &= 0b00011101;
  ind[6] &= 0b00011101;

  LOAD_displ();
  zaderj(DLITV);

  ind[0] &= 0b01100011;
  ind[2] &= 0b01100011;
  ind[4] &= 0b01100011;
  ind[6] &= 0b01100011;

  if (0b00000001 & ind[0]) ind[0] |= 0b00001000;
  if (0b00000001 & ind[2]) ind[2] |= 0b00001000;
  if (0b00000001 & ind[4]) ind[4] |= 0b00001000;
  if (0b00000001 & ind[6]) ind[6] |= 0b00001000;

  ind[0] &= 0b00001000;
  ind[2] &= 0b00001000;
  ind[4] &= 0b00001000;
  ind[6] &= 0b00001000;
  LOAD_displ();
  zaderj(DLITV);

  ind[0] = 0;
  ind[2] = 0;
  ind[4] = 0;
  ind[6] = 0;

  LOAD_displ();
  zaderj(DLITV);

  //загрузка

  if (0b00001000 & indb[0]) ind[0] |= 0b01000000;
  if (0b00001000 & indb[2]) ind[2] |= 0b01000000;
  if (0b00001000 & indb[4]) ind[4] |= 0b01000000;
  if (0b00001000 & indb[6]) ind[6] |= 0b01000000;

  LOAD_displ();
  zaderj(DLITV);

  ind[0] = 0;
  ind[2] = 0;
  ind[4] = 0;
  ind[6] = 0;

  if (0b00001000 & indb[0]) ind[0] |= 0b00000001;
  if (0b00001000 & indb[2]) ind[2] |= 0b00000001;
  if (0b00001000 & indb[4]) ind[4] |= 0b00000001;
  if (0b00001000 & indb[6]) ind[6] |= 0b00000001;

  if (0b00010000 & indb[0]) ind[0] |= 0b00100000;
  if (0b00010000 & indb[2]) ind[2] |= 0b00100000;
  if (0b00010000 & indb[4]) ind[4] |= 0b00100000;
  if (0b00010000 & indb[6]) ind[6] |= 0b00100000;

  if (0b00000100 & indb[0]) ind[0] |= 0b00000010;
  if (0b00000100 & indb[2]) ind[2] |= 0b00000010;
  if (0b00000100 & indb[4]) ind[4] |= 0b00000010;
  if (0b00000100 & indb[6]) ind[6] |= 0b00000010;

  if (0b00000001 & indb[0]) ind[0] |= 0b01000000;
  if (0b00000001 & indb[2]) ind[2] |= 0b01000000;
  if (0b00000001 & indb[4]) ind[4] |= 0b01000000;
  if (0b00000001 & indb[6]) ind[6] |= 0b01000000;

  LOAD_displ();
  zaderj(DLITV);

  ind[0] = indb[0];
  ind[2] = indb[2];
  ind[4] = indb[4];
  ind[6] = indb[6];
  LOAD_displ();
  zaderj(DLITV);
}
/*  выезд буфера снизу вверх */
void MAX7221::go_up(void)
{
  //очистка
  ind[0] &= 0b00011101;
  ind[2] &= 0b00011101;
  ind[4] &= 0b00011101;
  ind[6] &= 0b00011101;

  if (0b00000001 & ind[0]) ind[0] |= 0b01000000;
  if (0b00000001 & ind[2]) ind[2] |= 0b01000000;
  if (0b00000001 & ind[4]) ind[4] |= 0b01000000;
  if (0b00000001 & ind[6]) ind[6] |= 0b01000000;

  ind[0] &= 0b01111110;
  ind[2] &= 0b01111110;
  ind[4] &= 0b01111110;
  ind[6] &= 0b01111110;

  if (0b00010000 & ind[0]) ind[0] |= 0b00100000;
  if (0b00010000 & ind[2]) ind[2] |= 0b00100000;
  if (0b00010000 & ind[4]) ind[4] |= 0b00100000;
  if (0b00010000 & ind[6]) ind[6] |= 0b00100000;

  if (0b00000100 & ind[0]) ind[0] |= 0b00000010;
  if (0b00000100 & ind[2]) ind[2] |= 0b00000010;
  if (0b00000100 & ind[4]) ind[4] |= 0b00000010;
  if (0b00000100 & ind[6]) ind[6] |= 0b00000010;

  if (0b00001000 & ind[0]) ind[0] |= 0b00000001;
  if (0b00001000 & ind[2]) ind[2] |= 0b00000001;
  if (0b00001000 & ind[4]) ind[4] |= 0b00000001;
  if (0b00001000 & ind[6]) ind[6] |= 0b00000001;

  ind[0] &= 0b01100011;
  ind[2] &= 0b01100011;
  ind[4] &= 0b01100011;
  ind[6] &= 0b01100011;

  LOAD_displ();
  zaderj(DLITV);

  ind[0] &= 0b00011101;
  ind[2] &= 0b00011101;
  ind[4] &= 0b00011101;
  ind[6] &= 0b00011101;

  if (0b00000001 & ind[0]) ind[0] |= 0b01000000;
  if (0b00000001 & ind[2]) ind[2] |= 0b01000000;
  if (0b00000001 & ind[4]) ind[4] |= 0b01000000;
  if (0b00000001 & ind[6]) ind[6] |= 0b01000000;

  ind[0] &= 0b01000000;
  ind[2] &= 0b01000000;
  ind[4] &= 0b01000000;
  ind[6] &= 0b01000000;


  LOAD_displ();
  zaderj(DLITV);

  ind[0] = 0;
  ind[2] = 0;
  ind[4] = 0;
  ind[6] = 0;

  LOAD_displ();
  zaderj(DLITV);

  //загрузка

  if (0b01000000 & indb[0]) ind[0] |= 0b00001000;
  if (0b01000000 & indb[2]) ind[2] |= 0b00001000;
  if (0b01000000 & indb[4]) ind[4] |= 0b00001000;
  if (0b01000000 & indb[6]) ind[6] |= 0b00001000;

  LOAD_displ();
  zaderj(DLITV);

  ind[0] = 0;
  ind[2] = 0;
  ind[4] = 0;
  ind[6] = 0;

  if (0b01000000 & indb[0]) ind[0] |= 0b00000001;
  if (0b01000000 & indb[2]) ind[2] |= 0b00000001;
  if (0b01000000 & indb[4]) ind[4] |= 0b00000001;
  if (0b01000000 & indb[6]) ind[6] |= 0b00000001;

  if (0b00100000 & indb[0]) ind[0] |= 0b00010000;
  if (0b00100000 & indb[2]) ind[2] |= 0b00010000;
  if (0b00100000 & indb[4]) ind[4] |= 0b00010000;
  if (0b00100000 & indb[6]) ind[6] |= 0b00010000;

  if (0b00000010 & indb[0]) ind[0] |= 0b00000100;
  if (0b00000010 & indb[2]) ind[2] |= 0b00000100;
  if (0b00000010 & indb[4]) ind[4] |= 0b00000100;
  if (0b00000010 & indb[6]) ind[6] |= 0b00000100;

  if (0b00000001 & indb[0]) ind[0] |= 0b00001000;
  if (0b00000001 & indb[2]) ind[2] |= 0b00001000;
  if (0b00000001 & indb[4]) ind[4] |= 0b00001000;
  if (0b00000001 & indb[6]) ind[6] |= 0b00001000;

  LOAD_displ();
  zaderj(DLITV);

  ind[0] = indb[0];
  ind[2] = indb[2];
  ind[4] = indb[4];
  ind[6] = indb[6];
  LOAD_displ();
  zaderj(DLITV);
}
#define NAEZD 3
/*  наложение слева направо */
void MAX7221::imposeR(void)
{
  ind[0] &= 0b11111001; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[0] |= (indb[0] & 0b00000110);
  ind[0] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[0] |= (indb[0] & 0b01001001);
  ind[0] &= 0b01001111;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[0] = indb[0];
  //--------------------------------------------------------------
  ind[2] &= 0b11111001; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[2] |= (indb[2] & 0b00000110);
  ind[2] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[2] |= (indb[2] & 0b01001001);
  ind[2] &= 0b01001111;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[2] = indb[2];
  //--------------------------------------------------------------
  ind[4] &= 0b11111001; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[4] |= (indb[4] & 0b00000110);
  ind[4] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[4] |= (indb[4] & 0b01001001);
  ind[4] &= 0b01001111;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[4] = indb[4];
  //--------------------------------------------------------------
  ind[6] &= 0b11111001; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[6] |= (indb[6] & 0b00000110);
  ind[6] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[6] |= (indb[6] & 0b01001001);
  ind[6] &= 0b01001111;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[6] = indb[6];
  LOAD_displ();
  zaderj(NAEZD);
}

/* наложение справа налево */
void MAX7221::imposeL(void)
{
  ind[6] &= 0b0100111; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[6] |= (indb[6] & 0b00110000);
  ind[6] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[6] |= (indb[6] & 0b01001001);
  ind[6] &= 0b11111001;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[6] = indb[6];
  //--------------------------------------------------------------
  ind[4] &= 0b0100111; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[4] |= (indb[4] & 0b00110000);
  ind[4] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[4] |= (indb[4] & 0b01001001);
  ind[4] &= 0b11111001;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[4] = indb[4];
  //--------------------------------------------------------------
  ind[2] &= 0b0100111; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[2] |= (indb[2] & 0b00110000);
  ind[2] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[2] |= (indb[2] & 0b01001001);
  ind[2] &= 0b11111001;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[2] = indb[2];
  //--------------------------------------------------------------
  ind[0] &= 0b0100111; // стирание 1
  LOAD_displ();
  zaderj(NAEZD);

  ind[0] |= (indb[0] & 0b00110000);
  ind[0] &= 0b10110110;  // стирание 2
  LOAD_displ();
  zaderj(NAEZD);

  ind[0] |= (indb[0] & 0b01001001);
  ind[0] &= 0b11111001;  // стирание 3
  LOAD_displ();
  zaderj(NAEZD);

  ind[0] = indb[0];
  LOAD_displ();
  zaderj(NAEZD);
}
/* эффект колеса на всех индикаторах */
void MAX7221::koleso(uint8_t pomt) {
  clear_DISPL();
  for (int i = 1; i < 2; i++) {
    ind[0] = 0b01000000;
    ind[2] = 0b01000000;
    ind[4] = 0b01000000;
    ind[6] = 0b01000000;
    LOAD_displ();
    zaderj(pomt);
    ind[0] = 0b00100000;
    ind[2] = 0b00100000;
    ind[4] = 0b00100000;
    ind[6] = 0b00100000;
    LOAD_displ();
    zaderj(pomt);
    ind[0] = 0b00010000;
    ind[2] = 0b00010000;
    ind[4] = 0b00010000;
    ind[6] = 0b00010000;
    LOAD_displ();
    zaderj(pomt);
    ind[0] = 0b00001000;
    ind[2] = 0b00001000;
    ind[4] = 0b00001000;
    ind[6] = 0b00001000;
    LOAD_displ();
    zaderj(pomt);
    ind[0] = 0b00000100;
    ind[2] = 0b00000100;
    ind[4] = 0b00000100;
    ind[6] = 0b00000100;
    LOAD_displ();
    zaderj(pomt);
    ind[0] = 0b00000010;
    ind[2] = 0b00000010;
    ind[4] = 0b00000010;
    ind[6] = 0b00000010;
    LOAD_displ();
    zaderj(pomt);
  }
}
/* эффект колеса при изменении одного разряда */
void MAX7221::effect(uint8_t pomt) {
  ind[j] &= 0b01000000;
  ind[j] &= 0b00100000;
  ind[j] &= 0b00010000;
  ind[j] &= 0b00001000;
  ind[j] &= 0b00000100;
  ind[j] &= 0b00000010;
  ind[j] &= 0b00000001;
  LOAD_displ();
  zaderj(pomt);
  for (int i = 1; i < 7; i++) {
    ind[j] = 0b01000000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] = 0b00100000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] = 0b00010000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] = 0b00001000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] = 0b00000100;
    LOAD_displ();
    zaderj(pomt);
    ind[j] = 0b00000010;
    LOAD_displ();
    zaderj(pomt);

    ind[j] &= 0b01000000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] &= 0b00100000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] &= 0b00010000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] &= 0b00001000;
    LOAD_displ();
    zaderj(pomt);
    ind[j] &= 0b00000100;
    LOAD_displ();
    zaderj(pomt);
    ind[j] &= 0b00000010;
    LOAD_displ();
    zaderj(pomt);
  }
}